// AudioIn.cs

using System;
using System.Text;
using System.IO;
using System.Collections;

using System.Runtime.InteropServices;

using AudioLowLevel;

namespace AudioIn
{
    // An internal class used to get audio buffers
    internal class BufferIn : IDisposable
    {
        private Wave.WAVEHDR myHdr;
        private GCHandle myHdrHandle;
        private byte[] myData;
        private GCHandle myDataHandle;
        private IntPtr myaif;

        private AudioCapture myain;

        public BufferIn(IntPtr aif, AudioCapture ain)
        {
            myaif = aif;
            myain = ain;
        }

        public bool Init(int size)
        {
            // Allocate memory for myself
            myHdr = new Wave.WAVEHDR();
            myHdrHandle = GCHandle.Alloc(myHdr, GCHandleType.Pinned);

            myData = new byte[size];
            myDataHandle = GCHandle.Alloc(myData, GCHandleType.Pinned);

            // Pass the reference of the AudioIn caller
            myHdr.dwUser = (IntPtr)GCHandle.Alloc(myain);

            myHdr.lpData = myDataHandle.AddrOfPinnedObject();
            myHdr.dwBufferLength = size;

            // Prepare the headers
            Wave.waveInPrepareHeader(myaif, ref myHdr, (uint)Marshal.SizeOf(myHdr));

            // Put the buffer in the queue
            if (Wave.waveInAddBuffer(myaif, ref myHdr, (uint)Marshal.SizeOf(myHdr)) != 0)
                return false;

            return true;
        }

        public void Dispose()
        {
            Wave.waveInUnprepareHeader(myaif, ref myHdr, (uint)Marshal.SizeOf(myHdr));
        }

        internal static void WaveInProc(IntPtr hdrvr, int uMsg, int dwUser, ref Wave.WAVEHDR wavhdr, int dwParam2)
        {
            if (uMsg == Wave.MM_WIM_DATA)
            {
                try
                {
                    // Get a reference of the AudioIn class from the WAVEHDR
                    GCHandle h = (GCHandle)wavhdr.dwUser;
                    AudioCapture ain = (AudioCapture)h.Target;
                    ain.BlockDone();
                }
                catch
                {
                }
            }
        }

        public int GetData(ref IntPtr data)
        {
            data = myHdr.lpData;
            return (int)myHdr.dwBufferLength;
        }
    } // BufferIn ends



    // IDisposable so it can free unmanaged memory allocated
    public class AudioCapture : IDisposable
    {
        // Reference to audio interface
        private IntPtr aif = IntPtr.Zero;
        private Wave.PCMWAVEFORMAT PcmFormat;
        // Buffers used to store recording information.
        private BufferIn[] BuffersIn = null;
        private int curBlock;
        // Maximum number of buffers in the array
        private int nBlocks;
        // Size of each record buffer
        private uint bufferSize;

        // Initialization flag
        private bool inited = false;
        // Specifies whether or not recording is done
        private bool isRecording = false;

        private Wave.DoneCallback myBufferProc;

        // Delegate to call my caller passing him the audio buffer
        public delegate void BufferReadyDelegate(byte[] buffer, int size);
        private BufferReadyDelegate myBufferReady;

        public AudioCapture(BufferReadyDelegate nBufferReady)
        {
            this.myBufferReady = nBufferReady;
            this.nBlocks = 10;            
        }

        // Preload the buffers and prepare them for recording (nBufferSize must be a multiple of 2)
        public bool Load(int curDevice, short nSamples, short nBits, uint nBufferSize)
        {
            // If already registering
            if (isRecording)
                return false;

            // If already initialized start over
            if (inited)
                return true;

            bufferSize = nBufferSize;
            myBufferProc = new Wave.DoneCallback(BufferIn.WaveInProc);
            PcmFormat = new Wave.PCMWAVEFORMAT(nSamples, nBits);

            // Attempt to open the specified device with the desired wave format
            if (Wave.waveInOpen(ref aif, curDevice, ref PcmFormat, myBufferProc, 0, Wave.CALLBACK_FUNCTION) != 0)
                return false;

            // Allocate the array of buffers
            BuffersIn = new BufferIn[nBlocks + 1];

            // Initialize two buffers to start with
            BuffersIn[0] = new BufferIn(aif, this);
            BuffersIn[1] = new BufferIn(aif, this);

            if (!InitBuffer(0) || !InitBuffer(1))
                return false;

            curBlock = 0;
            inited = true;

            return true;
        }

        // Initialize a buffer for recording by allocating a data buffer, preparing the header, and adding it to the read queue.
        public bool InitBuffer(int bufIndex)
        {
            // If the header is not already instanced then instance it
            if (BuffersIn[bufIndex] == null)
                BuffersIn[bufIndex] = new BufferIn(aif, this);

            // Allocate memory if not already allocated 
            if (!BuffersIn[bufIndex].Init((int)bufferSize))
                return false;

            return true;
        }

        // Called from the callback, it means the block is recorded
        public void BlockDone()
        {
            // cycle throught the array
            short nextBlock = (short)((this.curBlock == nBlocks) ? 0 : this.curBlock + 1);

            InitBuffer(nextBlock);

            // Get the data recorded
            IntPtr dataPnt = IntPtr.Zero;
            int size = 0;
            size = BuffersIn[curBlock].GetData(ref dataPnt);

            // Copy the data pointed to a managed byte[]
            byte[] data = new Byte[size];
            System.Runtime.InteropServices.Marshal.Copy(dataPnt, data, 0, size);

            // Free headers
            BuffersIn[curBlock].Dispose();

            // The buffer ready callback to AudioMuxer
            this.myBufferReady(data, size);

            this.curBlock = nextBlock;
        }

        // Start recording
        public bool Start()
        {
            // If not yet ready or already started
            if (!inited || isRecording)
                return false;

            if (Wave.waveInStart(aif) != 0)
                return false;

            isRecording = true;
            return true;
        }

        // Stop recording
        public void Stop()
        {
            Wave.waveInStop(aif);

            FreeBuffers();
            isRecording = false;
        }

        // Free the buffers
        public void Dispose()
        {
            myBufferProc = null;
            FreeBuffers();
        }

        // Free the buffers allocated for recording
        private void FreeBuffers()
        {
            inited = false;

            if (BuffersIn != null)
            {
                for (int i = 0; i < BuffersIn.Length; i++)
                {
                    if (BuffersIn[i] != null)
                    {
                        BuffersIn[i].Dispose();
                        BuffersIn[i] = null;
                    }
                }
                BuffersIn = null;
            }

            Wave.waveInClose(aif);
            aif = IntPtr.Zero;
        }

    } // end of AudioCapture


}