// AudioOut.cs

using System.Threading;

using System;
using System.Text;
using System.IO;
using System.Collections;
using System.Runtime.InteropServices;

using AudioLowLevel;

namespace AudioOut
{

    // An internal class used to fill audio buffers
    internal class BufferOut : IDisposable
    {
        private Wave.WAVEHDR myHdr; 

        private GCHandle myHdrHandle;
        private byte[] myData;
        private GCHandle myDataHandle;
        private IntPtr myaif;

        public BufferOut(IntPtr aif, AudioPlayer aout, byte[] aData, int aDataSize)
        {
            myaif = aif;
        
            // Allocate memory for myself
            myHdrHandle = GCHandle.Alloc(myHdr, GCHandleType.Pinned);
            // Pass the reference of the AudioOut caller
            myHdr.dwUser = (IntPtr)GCHandle.Alloc(aout);

            myData = new byte[aDataSize];
            myDataHandle = GCHandle.Alloc(myData, GCHandleType.Pinned);
            myHdr.lpData = myDataHandle.AddrOfPinnedObject();
            myHdr.dwBufferLength = aDataSize;
            
            // Prepare the headers
            uint res = Wave.waveOutPrepareHeader(myaif, ref myHdr, (uint)Marshal.SizeOf(myHdr));

            // copy audio data to the buffer
            Marshal.Copy(aData, 0, myHdr.lpData, aDataSize);

            // Write the buffer to system queue
            Wave.waveOutWrite(myaif, ref myHdr, (uint)Marshal.SizeOf(myHdr));
        }

        public void Dispose()
        {
            Wave.waveOutUnprepareHeader(myaif, ref myHdr, (uint)Marshal.SizeOf(myHdr));
        }

        internal static void WaveOutProc(IntPtr hdrvr, int uMsg, int dwUser, ref Wave.WAVEHDR wavhdr, int dwParam2)
        {
            if (uMsg == Wave.MM_WOM_DONE)
            {
                try
                {
                    // Get a reference of the AudioIn class from the WAVEHDR
                    GCHandle h = (GCHandle)wavhdr.dwUser;
                    AudioPlayer aout = (AudioPlayer)h.Target;
                    aout.BlockDone();
                }
                catch
                {
                }
            }
        }
    } // BufferOut ends


    // IDisposable so it can free unmanaged memory allocated
    public class AudioPlayer : IDisposable
    {

        // Reference to audio interface
        private IntPtr aif = IntPtr.Zero;
        private Wave.PCMWAVEFORMAT PcmFormat;
        // Buffers used to store recording information.
        private BufferOut[] BuffersOut = null;
        private short curBuf;
        private bool isPlaying;
        private int curDevice;

        // Delegate from the bufferout to audioplayer notifing buffer is done
        private Wave.DoneCallback myBufferProc;

        // Delegate to call audiomuxer requesting an audio data
        public delegate void BufferDoneDelegate(short curBuf);
        private BufferDoneDelegate myBufferDone;

        public AudioPlayer(int curDevice, short nSamples, short nBits, BufferDoneDelegate nBufferDone)
        {
            this.curDevice = curDevice;
            this.myBufferDone = nBufferDone;
            this.isPlaying = false;

            this.myBufferProc = new Wave.DoneCallback(BufferOut.WaveOutProc);
            this.PcmFormat = new Wave.PCMWAVEFORMAT(nSamples, nBits);

            // I keep two buffer locally for fast switching 
            this.BuffersOut = new BufferOut[2];
            this.curBuf = 0;
        }

        public bool Load()
        {
            // If already playing
            if (isPlaying)
                return false;

            // Attempt to open the specified device with the desired wave format
            if (Wave.waveOutOpen(ref this.aif, this.curDevice, ref this.PcmFormat, this.myBufferProc, 0, Wave.CALLBACK_FUNCTION) != 0)
                return false;
           
            curBuf = 0;
            isPlaying = true;

            return true;
        }

        // called from AudioDemuxer, fills values
        public void InitBuffer(short bufNum, byte[] aData, int aDataSize)
        {
            this.BuffersOut[bufNum] = new BufferOut(this.aif, this, aData, aDataSize);
        }
          
        // Called from the callback, it means the block is played
        public bool BlockDone()
        {
            short nextBuf = (short)((this.curBuf == 0) ? 1 : 0);

            // clear the done buffer
            this.BuffersOut[curBuf].Dispose();

            Thread.Sleep(50);

            // fill the done buffer with next audio data calling the callback
            this.myBufferDone(curBuf);

            this.curBuf = nextBuf;

            return true;
        }
    

        // Start playing
        public bool Play()
        {
            if (isPlaying)
                return true;

            if (Wave.waveOutRestart(aif) != 0)
               return false;

            isPlaying = true;
            return true;
        }

        // Stop playing
        public void Pause()
        {
            Wave.waveOutPause(aif);

            isPlaying = false;
        }

        // Free the buffers
        public void Dispose()
        {
            myBufferProc = null;
            FreeBuffers();
        }

        // Free the buffers allocated for recording
        private void FreeBuffers()
        {   
            if (BuffersOut != null)
            {
                for (int i = 0; i < BuffersOut.Length; i++)
                {
                    if (BuffersOut[i] != null)
                    {
                        BuffersOut[i].Dispose();
                        BuffersOut[i] = null;
                    }
                }
                BuffersOut = null;
            }

            Wave.waveOutClose(aif);
            aif = IntPtr.Zero;
        }

    } // end of AudioPlayer


}