// Demuxer.cs

using System;
using System.Text;
using System.IO;
using System.Drawing;
using System.Drawing.Imaging;
using System.Threading;

using System.Collections;

using AudioOut;
using Network;

namespace Mux
{

    // AudioDemuxer get data from network packet
    public class AudioDemuxer
    {
        private int currentDevice;
        private int playoutDelay;

        private bool isRunning = false;

        private Queue syncQueue;
        private PacketList pList;
        private NetClient netClient;
        private AudioPlayer aout;
        private Thread netThr;
        private Thread decodeThr;

        // Callback from audioplayer
        private AudioPlayer.BufferDoneDelegate bufferDone;

        public AudioDemuxer(Queue tsqueue, string addr, int port, int nAudioDev, short nSamples, short nBits, int playoutDelay)
        {

            // Initialization
            this.currentDevice = nAudioDev;
            this.playoutDelay = playoutDelay;
            this.pList = new PacketList();
            this.syncQueue = tsqueue;

            this.bufferDone = new AudioPlayer.BufferDoneDelegate(this.RequestData);

            this.aout = new AudioPlayer(nAudioDev, nSamples, nBits, this.bufferDone);
            this.netClient = new NetClient(addr, port);

            // Starts network-related thread
            this.isRunning = true;
            netThr = new Thread(new ThreadStart(this.NetThread));
            netThr.Start();

            // Thread managing audio player. Needed for sleeping stuff
            decodeThr = new Thread(new ThreadStart(this.DecodePacket));
            decodeThr.Start();
        
        }

        public void Stop()
        {
            this.isRunning = false;

            this.netClient.Close();
            this.netThr.Join();
            this.decodeThr.Join();
        }

        #region Read network data and fills buffer
        public void NetThread()
        {
            byte[] pktByte = null;
            int lenght = 0;

            while (this.isRunning)
            {
                this.netClient.recvByte(ref pktByte, ref lenght);

                if ((lenght != pktByte.Length) || (lenght == 0))
                    continue;
//                    throw new Exception("Bad packet length received.");

                // Rebuild the packet from the buffer, if it`s ok add to the list
                AUDIOPACKET aPacket = new AUDIOPACKET();
                if (!aPacket.SetBytes(pktByte))
                {
                    // Add the packet to the list
                    this.pList.Add(aPacket);
                }
            }
        }
        #endregion

        public void DecodePacket()
        {
            while ((this.pList.Size() < 2) && isRunning)
                Thread.Sleep(playoutDelay);

            if (this.isRunning)
            {
                if (!this.aout.Load())
                    throw new Exception("Loading audio player failed.");

                // load first buffers in audioOut
                AUDIOPACKET myap = this.pList.Get();
                this.aout.InitBuffer(0, myap.AudioData, myap.AudioDataSize);
                myap = this.pList.Get();
                this.aout.InitBuffer(1, myap.AudioData, myap.AudioDataSize);
            }
        }

        // Called from AudioPlayer to request next buffer filling
        public void RequestData(short curBuf)
        {
            if (!isRunning)
                return;

            lock (this.pList)
            {

                if (this.pList.Size() < 2)
                {
                    // I can`t play without enough packet, pause the player 
                    this.aout.Pause();
                    this.DecodePacket();
                }
                else
                {
                    // Get the first inserted packet
                    AUDIOPACKET myap = this.pList.Get();

                    // load the buffer in audioOut
                    this.aout.InitBuffer(curBuf, myap.AudioData, myap.AudioDataSize);
                    // maybe i was in pause
                    this.aout.Play();

                    // send the timestamp and cursor info to videoDemuxer
                    QueueObj qo = new QueueObj(myap.timeStamp, myap.mCursorSize, myap.mCursor, myap.flags, myap.xMouse, myap.yMouse);
                    this.syncQueue.Enqueue(qo);
                    Thread.Sleep(0);
                }
            }
        }
    }    

    // Manage the list of packets (is an Hashtable)
    public class PacketList
    {
        private Hashtable ht;

        public PacketList()
        {
            // The hastable is synchronized, ie. removes of main thread not interfere with netThr
            ht = Hashtable.Synchronized(new Hashtable());
        }

        public void Add(AUDIOPACKET value){
            lock (ht)
            {
                try
                {
                    ht.Add(value.timeStamp, value);
                }
                catch
                {
                }
            }
        }

        // Returns the first packet and remove it from list
        public AUDIOPACKET Get(){
            long firstKey = long.MaxValue;
            AUDIOPACKET ap;

            lock (ht)
            {
                ICollection keyColl = ht.Keys;

                foreach (long key in keyColl)
                    if (firstKey > key)
                        firstKey = key;

                ap = (AUDIOPACKET) ht[firstKey];
                ht.Remove(firstKey);
            }
            return ap;

        }

        // Returns number of packets present
        public int Size()
        {
            return ht.Count;
        }
    }

}
