
using System;
using System.Threading;
using System.Collections;
using System.Drawing;
using System.IO;

using Network;
using Video;

namespace Mux
{
    // Gets video packets and sincronize presentation with audio flux
    public class VideoDemuxer
    {
        private VideoPlayer vPlayer;
        private NetClient netClient;
        private Queue syncQueue;

        private Thread gpThr;
        private Thread syncThr;

        private bool isRunning;

        public VideoDemuxer(Queue tsqueue, string addr, int port)
        {
            // init
            this.vPlayer = new VideoPlayer();
            this.netClient = new NetClient(addr, port);
            this.syncQueue = tsqueue;

            // Starts network-related thread
            this.isRunning = true;
            gpThr = new Thread(new ThreadStart(this.GetPacketThread));
            gpThr.Start();

            // Thread managing video synchronization
            syncThr = new Thread(new ThreadStart(this.SyncThread));
            syncThr.Start();

            this.vPlayer.Show();
        }

        public void Stop()
        {
            this.isRunning = false;
            this.vPlayer.isRunning = false;

            if (vPlayer != null)
            {
                this.netClient.Close();
                this.vPlayer.Close();

                this.gpThr.Join();
                this.syncThr.Join();
            }
        }

        private void GetPacketThread()
        {
            byte[] pktByte = null;
            int length = 0;

            while (this.isRunning)
            {
                this.netClient.recvByte(ref pktByte, ref length);

                if ((length != pktByte.Length) || (length == 0))
                    continue;

                // Rebuild the packet from the buffer, if it`s ok send to VideoPlayer
                VIDEOPACKET vPacket = new VIDEOPACKET();
                if (vPacket.SetBytes(pktByte))
                    if (vPacket.frameDataSize != 0)
                        if (this.vPlayer.isRunning)
                            this.vPlayer.Invoke(this.vPlayer.afDeleg, vPacket.timeStamp, vPacket.frameData, vPacket.isKeyframe());
            }
        }

        // controls when the next frame has to be drawn, call the delegate
        private void SyncThread()
        {
            long tsLast = 0;
            long tsCurr = -1; 
            uint tsDiff;
            int deltatsDiff = 0;
            
            QueueObj queueObj;
            long tsToFollow = 0;

            bool mDraw = false;
            byte[] bCursor = new byte[0];
            int bCursorSize = 0;
            ushort mx = 0;
            ushort my = 0;

            while (this.isRunning)
            {
                // show the next frame
                if (this.vPlayer.isRunning)
                    tsCurr = (long)this.vPlayer.Invoke(this.vPlayer.sfDeleg, mDraw, bCursorSize, bCursor, mx, my);

                // no frames available, wait some time, clean the queue with already present timestamps
                if (tsCurr == -1)
                {
                    Thread.Sleep(3000);
                    this.syncQueue.Clear();
                }
                else
                {
                    // get the first timestamp waiting
                    if (this.syncQueue.Count > 0)
                    {
                        queueObj = (QueueObj)this.syncQueue.Dequeue();
                        tsToFollow = queueObj.GetInfo(ref mDraw, ref bCursorSize, ref bCursor, ref mx, ref my);

                        // If video fps is too much lower than audioblock size I go out of sync, so drop some timing info
                        if (this.syncQueue.Count > 5)
                           this.syncQueue.Clear();                            
                    }

                    // it`s the first packet, I don`t know sleeping time
                    if (tsLast == 0)
                        tsDiff = 0;
                    else
                    {
                        // Basically the time to sleep is the gap between two video frame 
                        tsDiff = (uint)(tsCurr - tsLast);

                        // I have the audio sync information?
                        if (tsToFollow != 0)
                        {
                            // The delta if is not too late nor way sooner. Can have approx half second of desync
                  //          deltatsDiff = 0;

                            // Video is too late
                            if (tsCurr <= tsToFollow)
                                deltatsDiff -= 4000;

                            // The video is way sooner than audio (the next frame it will also be sooner)
                            if ((tsCurr + tsDiff) > tsToFollow)
                                deltatsDiff += 4000; // Slow down the video 
                            
                            tsToFollow = 0;
                        }
                    }
                    
                    Thread.Sleep((int)((uint)(tsDiff + deltatsDiff) / 10000)); // convert from filetime to milliseconds
                    
                    tsLast = tsCurr;
                }
            }
        }

    }
}
