// Packets.cs : packets definitions

using System;
using System.Drawing;
using System.Drawing.Imaging;
using System.Security.Cryptography;
using System.IO;

namespace Network
{
 
    // Audio packet struct
    public struct AUDIOPACKET
    {
        public long timeStamp; // 8 byte

        public int AudioDataSize; // 4 byte
        public byte[] AudioData;

        public ushort xMouse; // 2 byte - X coordinate of mouse pointer
        public ushort yMouse; // 2 byte - Y coordinate of mouse pointer
        public byte flags;
        public ushort mCursorSize; // 2 byte
        public byte[] mCursor; // The mouse icon

        private byte[] CRC; // 16 byte

        // Sets the values and return the total size
        public AUDIOPACKET(DateTime now, int nAudioDataSize, byte[] nAudioData, 
            Point pntMouse, ushort nMouseDataSize, byte[] nMouseData, bool nDrawCursor)
        {
            timeStamp = now.ToFileTimeUtc();

            AudioDataSize = nAudioDataSize;
            AudioData = nAudioData;

            xMouse = (ushort)pntMouse.X;
            yMouse = (ushort)pntMouse.Y;

            mCursorSize = nMouseDataSize;
            mCursor = nMouseData;

            // First bit is true/false if the cursor has to be drawn 
            flags = 0;
            if (nDrawCursor)
                flags |= 0x1;

            // Checksum is MD5 (16 bytes)
            CRC = new byte[16];
        }

        // Build the byte sequence
        public byte[] GetBytes()
        {
            MemoryStream ms = new MemoryStream();
           
            ms.Write(BitConverter.GetBytes(timeStamp), 0, 8);
            ms.Write(BitConverter.GetBytes(AudioDataSize), 0, 4);
            ms.Write(AudioData, 0, AudioDataSize);
            ms.Write(BitConverter.GetBytes(xMouse), 0, 2);
            ms.Write(BitConverter.GetBytes(yMouse), 0, 2);
            ms.WriteByte(flags);
            ms.Write(BitConverter.GetBytes(mCursorSize), 0, 2); 
            ms.Write(mCursor, 0, mCursorSize); 
            
            // Calculate and set the checksum
            CRC = MD5.Create().ComputeHash(ms.ToArray());
            ms.Write(CRC, 0, 16);

            return ms.ToArray();
        }

        // Fill values from a byte sequence, returns true if bad data
        public bool SetBytes(byte[] pktBytes)
        {
            if (pktBytes.Length == 0)
                return true;

            // I have to catch exception from bad indexes
            try
            {
                MemoryStream ms = new MemoryStream(pktBytes);
                byte[] tmpBytes = new byte[pktBytes.Length];
                
                CRC = new byte[16];

                ms.Read(tmpBytes, 0, 8);
                timeStamp = (long)BitConverter.ToInt64(tmpBytes, 0);
                
                ms.Read(tmpBytes, 0, 4);
                AudioDataSize = (int)BitConverter.ToInt32(tmpBytes, 0);
                AudioData = new byte[AudioDataSize];
                ms.Read(AudioData, 0, AudioDataSize);
                
                ms.Read(tmpBytes, 0, 2);
                xMouse = (ushort)BitConverter.ToUInt16(tmpBytes, 0);
                ms.Read(tmpBytes, 0, 2);
                yMouse = (ushort)BitConverter.ToUInt16(tmpBytes, 0);
                
                flags = (byte)ms.ReadByte();
                
                ms.Read(tmpBytes, 0, 2);
                mCursorSize = (ushort)BitConverter.ToUInt16(tmpBytes, 0);
                mCursor = new byte[mCursorSize];
                ms.Read(mCursor, 0, mCursorSize);
                
                ms.Read(CRC, 0, 16);

                // Calculate the CRC, and compare        
                byte[] nCRC = MD5.Create().ComputeHash(ms.ToArray(), 0, (int)ms.Length - 16);

                return !this.checkCRC(CRC, nCRC);
            }
            catch (System.ArgumentOutOfRangeException)
            {
                return true;
            }
        }

        private bool checkCRC(byte[] crc0, byte[] crc1)
        {
            bool val = false;

            for (uint i = 0; i < crc0.Length; i++)
                val |= crc0[i] == crc1[i];

            return val;
        }
    }



    // Video packet struct
    public struct VIDEOPACKET
    {
        public long timeStamp; // 8 byte

        public int frameDataSize; // 4 byte
        public byte[] frameData;
        public byte flags;

        private byte[] CRC; // 16 byte

        public VIDEOPACKET(DateTime now, int size, byte[] buffer, bool tof)
        { 
            timeStamp = now.ToFileTimeUtc();

            frameDataSize = size;
            frameData = buffer;

            // tof is true if it`s a keyframe
            flags = 0x0;
            if (!tof)
                flags |= 0x1; 

            // Checksum is MD5 (16 bytes)
            CRC = new byte[16];
        }

        // Build the byte sequence
        public byte[] GetBytes()
        {
            MemoryStream ms = new MemoryStream();

            ms.Write(BitConverter.GetBytes(timeStamp), 0, 8);
            ms.Write(BitConverter.GetBytes(frameDataSize), 0, 4);
            ms.Write(frameData, 0, frameDataSize);
            ms.WriteByte(flags);

            // Calculate and set the checksum
            CRC = MD5.Create().ComputeHash(ms.ToArray());
            ms.Write(CRC, 0, 16);

            return (byte[]) ms.ToArray();
        }

        // Fill values from a byte sequence
        public bool SetBytes(byte[] pktBytes)
        {
            if (pktBytes.Length == 0)
                return true;

            // I have to catch exception from bad indexes
            try
            {
                MemoryStream ms = new MemoryStream(pktBytes);

                byte[] tmpBytes = new byte[8];
                frameData = new byte[pktBytes.Length];
                CRC = new byte[16];

                ms.Read(tmpBytes, 0, 8);
                timeStamp = (long)BitConverter.ToInt64(tmpBytes, 0);
                
                ms.Read(tmpBytes, 0, 4);
                frameDataSize = (int)BitConverter.ToInt32(tmpBytes, 0);
                ms.Read(frameData, 0, frameDataSize);

                flags = (byte)ms.ReadByte();

                ms.Read(CRC, 0, 16);

                // Calculate the CRC, and compare        
                byte[] nCRC = MD5.Create().ComputeHash(ms.ToArray(), 0, (int)ms.Length - 16);

                return this.checkCRC(CRC, nCRC);
            }
            catch (System.ArgumentOutOfRangeException)
            {
                return true;
            }
        }

        // returns true if correct
        private bool checkCRC(byte[] crc0, byte[] crc1)
        {
            bool val = false;

            for (uint i = 0; i < crc0.Length; i++)
                val |= crc0[i] == crc1[i];

            return val;
        }

        public bool isKeyframe()
        {
            return (flags == 0x1) ? false : true;
        }
    }

}
