using System;
using System.Drawing;
using System.Drawing.Imaging;
using System.Drawing.Design;

using System.IO;
using System.IO.Compression;

namespace Video
{
    class VideoCompression
    {
        
        // returns a difference stream compressed. The width and height must be a multiple of 16 pixels
        public static byte[] GetDiff(Bitmap curBmp, Rectangle rect, Bitmap lastKeyBmp)
        {
            // array containing byte values of bmp blocks 
            MemoryStream bmpDiff = new MemoryStream();

            // lock the image memory region
            BitmapData lastBmpData = lastKeyBmp.LockBits(rect, ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb);
            BitmapData curBmpData = curBmp.LockBits(rect, ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb);

            IntPtr lScan0 = lastBmpData.Scan0;
            IntPtr cScan0 = curBmpData.Scan0;

            // the block count of a row and a column
            // because of RGB, the matrix block is 16*3 x 16 = 48x16
            int blocksW = lastBmpData.Stride / 48;
            int blocksH = lastKeyBmp.Height / 16;
            
            // temp repository
            byte[] block = new byte[768]; // 48*16 pixels
            int cnt = 0;
            int l = 0;
            int index;

            // working with pointer is faster, compile with /unsafe
            unsafe
            {
                byte* lp = (byte*)(void*)lScan0;
                byte* cp = (byte*)(void*)cScan0;

                for (int x = 0; x < blocksH; x++)
                {
                    for (int y = 0; y < blocksW; y++)
                    {
                        // works on current block
                        for (int j = 0; j < 16; j++)
                        {
                            for (int i = 0; i < 48; i++)
                            {
                                index = i + (j * lastBmpData.Stride) + (y * 48);
                                if ((cp[index] - lp[index]) != 0)
                                    cnt++;

                                block[l++] = cp[index];
                            }
                        }

                        // XXX if more of 96 bytes changed, keep the difference block
                        if (cnt > 96)
                        {
                           // add the number of block and the diff block to the vector
                            bmpDiff.Write(BitConverter.GetBytes((ushort)(x*blocksW + y)), 0, 2);
                            bmpDiff.Write(block, 0, 768);
                        }
                        cnt = l = 0;
                    }
                    lp += lastBmpData.Stride * 16;
                    cp += lastBmpData.Stride * 16;
                }
            }

            // unlock the image memory region
            lastKeyBmp.UnlockBits(lastBmpData);
            curBmp.UnlockBits(curBmpData);

            return VideoCompression.gzip(bmpDiff.ToArray());
        }


        // merge the difference stream with the last known bitmap
        public static void GetBmp(byte[] cdiff, Rectangle rect, ref Bitmap lastBmp)
        {
            // the lenght is 0, so the frame is identical
            if (cdiff.Length == 0)
                return;
            byte[] diff = VideoCompression.gunzip(cdiff);
            if (diff.Length == 0)
                return;
            
            // array containing byte values of bmp blocks 
            MemoryStream bmpDiff = new MemoryStream(diff);

            // lock the image memory region
            BitmapData lastBmpData = lastBmp.LockBits(rect, ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb);

            IntPtr lScan0 = lastBmpData.Scan0;

            // the block count of a row and a column
            // because of RGB, the matrix block is 16*3=48     
            int blocksW = lastBmpData.Stride / 48;
            int blocksH = lastBmp.Height / 16; // *3/48

            // read the first block position to update
            byte[] posByte = new byte[2];
            ushort position;
            byte[] block = new byte[768]; // 48*16 pixels              
            int l, index;

            bmpDiff.Read(posByte, 0, 2);
            position = (ushort)BitConverter.ToUInt16(posByte, 0);
            bmpDiff.Read(block, 0, 768);

            // working with pointer is faster, compile with /unsafe
            unsafe
            {
                byte* lp = (byte*)(void*)lScan0;


                for (int x = 0; x < blocksH; x++)
                {
                    for (int y = 0; y < blocksW; y++)
                    {
                        if (position == (x * blocksW + y))
                        {
                            l = 0;
                            // works on current block
                            for (int j = 0; j < 16; j++)
                            {
                                for (int i = 0; i < 48; i++)
                                {
                                    index = i + (j * lastBmpData.Stride) + (y * 48);
                                    lp[index] = (byte)(block[l++]);
                                }
                            }

                            // it was the last block to update, so quit
                            if (bmpDiff.Position == bmpDiff.Length)
                                break;

                            bmpDiff.Read(posByte, 0, 2);
                            position = (ushort)BitConverter.ToUInt16(posByte, 0);
                            bmpDiff.Read(block, 0, 768);

                        }
                    }
                    lp += lastBmpData.Stride * 16;
                }                
            }
            // unlock the image memory region
            lastBmp.UnlockBits(lastBmpData);
        }

        // simple function for fast gzip compression
        public static byte[] gzip(byte[] data)
        {
            MemoryStream ms = new MemoryStream();
            Stream s = new GZipStream(ms, CompressionMode.Compress);
            s.Write(data, 0, data.Length);
            s.Close();

            return (byte[])ms.ToArray();
        }

        public static byte[] gunzip(byte[] data)
        {
            MemoryStream ms = new MemoryStream(data);
            Stream s = new GZipStream(ms, CompressionMode.Decompress);

            byte[] buff = new byte[data.Length * 20];

            try
            {
                s.Read(buff, 0, buff.Length);
                s.Close();
            }
            catch (Exception)
            {
                return new byte[0];
            }
            return (byte[])buff;
        }
    }
}
